"""The Material `Stylesheet` can be included in your app
if you want to have customized styles for your widgets and panes.

The styles are based on [Material Design](https://material.io/design) and the
[MWC](https://github.com/material-components/material-components-web-components) implementation.

As an example you can set the `primary_color` and this corresponds to the `--mdc-theme-primary`
css value.

The Stylesheet `.editor` parameter provides an interactive stylesheet editor.
"""
import panel as pn
import param

_STYLE_PARAMETERS = [
    "primary_color",
    "primary_on_color",
    "secondary_color",
    "secondary_on_color",
    "error_color",
    "error_on_color",
]
_SETTINGS_PARAMETERS = _STYLE_PARAMETERS + ["reset_to_defaults"]
_STYLEPANE_RENAME = {p: None for p in _SETTINGS_PARAMETERS}


class Stylesheet(pn.pane.HTML):
    """The Material `Stylesheet` can be included in your app
if you want to have customized styles for your widgets and panes.

The styles are based on [Material Design](https://material.io/design) and the
[MWC](https://github.com/material-components/material-components-web-components) implementation.

As an example you can set the `primary_color` and this corresponds to the `--mdc-theme-primary`
css value.

The Stylesheet `.editor` parameter provides an interactive stylesheet editor.
"""

    primary_color = param.Color(default="#4caf50")
    primary_on_color = param.Color(default="#000000")

    secondary_color = param.Color(default="#9c27b0")
    secondary_on_color = param.Color(default="#ffffff")

    error_color = param.Color(default="#f44336")
    error_on_color = param.Color(default="#000000")

    reset_to_defaults = param.Action()

    editor = param.Parameter(doc="An interactive style editor component")

    # Needed for inheritance to work
    priority = 0
    _rename = {**pn.pane.HTML._rename, **_STYLEPANE_RENAME}

    def __init__(self, **params):
        params["height"] = 0
        params["width"] = 0
        params["sizing_mode"] = "fixed"
        params["margin"] = 0
        super().__init__(**params)

        self.editor = pn.WidgetBox(
            pn.Param(self, parameters=(_SETTINGS_PARAMETERS),), name="Material StyleSheet Editor"
        )
        self.reset_to_defaults = self._reset_to_defaults

        self._handle_style_parameter_change()

    # Don't name the function
    # `_update`, `_update_object_from_parameters`, `_update_model` or `_update_pane`
    # as this will override a function in the parent class.
    @param.depends(*_STYLE_PARAMETERS, watch=True)
    def _handle_style_parameter_change(self, *_):
        self.object = f"""
<style>
:root {{
  --mdc-theme-primary: {self.primary_color};
  --mdc-theme-on-primary: {self.primary_on_color};
  --mdc-theme-secondary: {self.secondary_color};
  --mdc-theme-on-secondary: {self.secondary_on_color};
  --mdc-theme-error: {self.error_color};
  --mdc-theme-on-error: {self.error_on_color};

  --mdc-typography-button-font-size: 1.33rem;
}}
body {{
    font-family: roboto;
    font-size: 14px;
}}
mwc-circular-progress {{
    margin-left: auto;
    margin-right: auto;
    display: block;
}}

mwc-button.secondary,
mwc-linear-progress.secondary,
mwc-circular-progress.secondary
{{
    --mdc-theme-primary: {self.secondary_color};
    --mdc-theme-on-primary: {self.secondary_on_color};
}}
mwc-button.warning, mwc-button.danger,
mwc-linear-progress.warning, mwc-linear-progress.danger,
mwc-circular-progress.warning, mwc-circular-progress.danger
{{
    --mdc-theme-primary: {self.error_color};
    --mdc-theme-on-primary: {self.error_on_color};
}}
mwc-button.light,
mwc-linear-progress.light,
mwc-circular-progress.light
{{
    --mdc-theme-primary: #fafafa;
    --mdc-theme-on-primary: black;
}}
mwc-button.dark,
mwc-linear-progress.dark,
mwc-circular-progress.dark
{{
    --mdc-theme-primary: #212121;
    --mdc-theme-on-primary: white;
}}
</style>
"""

    def _reset_to_defaults(self, *_):
        defaults = {p: self.param[p].default for p in _STYLE_PARAMETERS}
        self.param.set_param(**defaults)

    # Needed to avoid infinite recursion
    def __str__(self):
        return "Stylesheet()"

    def __repr__(self, depth=0):  # pylint: disable=unused-argument
        return "Stylesheet()"



from panel.pane import HTML
FAST_JS_MODULE = "https://unpkg.com/@microsoft/fast-components"
FAST_JS_SCRIPT = (
    f'<script type="module" src="{FAST_JS_MODULE}"></script>'
)

def get_fast_js_panel():
    return HTML(FAST_JS_SCRIPT, width=0, height=0, sizing_mode="fixed", margin=0)



"""Asset exports"""
ROOT_SPINNER_URL = (
    "https://github.com/MarcSkovMadsen/awesome-panel/raw/master/assets/images/spinners/"
)
SPINNER_AQUA = ROOT_SPINNER_URL + "spinner_aqua.png"
SPINNER_PANEL_BLINK_LIGHT_400_340 = ROOT_SPINNER_URL + "spinner_panel_blink_light_400_340.gif"
SPINNER_PANEL_BREATH_LIGHT_400_170 = ROOT_SPINNER_URL + "spinner_panel_breath_light_400_170.gif"
SPINNER_PANEL_BREATH_LIGHT_400_340 = ROOT_SPINNER_URL + "spinner_panel_breath_light_400_340.gif"
SPINNER_PANEL_GREEN_DARK = ROOT_SPINNER_URL + "spinner_panel_green_dark.gif"
SPINNER_PANEL_GREEN_DARK_STATIC = ROOT_SPINNER_URL + "spinner_panel_green_dark_static.gif"
SPINNER_PANEL_GREEN_LIGHT = ROOT_SPINNER_URL + "spinner_panel_green_light.gif"
SPINNER_PANEL_GREEN_LIGHT_STATIC = ROOT_SPINNER_URL + "spinner_panel_green_light_static.gif"
SPINNER_PANEL_ROTATE_400_400 = ROOT_SPINNER_URL + "spinner_panel_rotate_400_400.gif"
SPINNER_PANEL_STATIC_LIGHT = ROOT_SPINNER_URL + "spinner_panel_static_light.png"
SPINNER_PANEL_STATIC_LIGHT_400_340 = ROOT_SPINNER_URL + "spinner_panel_static_light_400_340.gif"



"""List of mwc-icon options"""
# source: https://github.com/google/material-design-icons/blob/master/iconfont/codepoints
MWC_ICONS = [
    None,
    "3d_rotation",
    "ac_unit",
    "access_alarm",
    "access_alarms",
    "access_time",
    "accessibility",
    "accessible",
    "account_balance",
    "account_balance_wallet",
    "account_box",
    "account_circle",
    "adb",
    "add",
    "add_a_photo",
    "add_alarm",
    "add_alert",
    "add_box",
    "add_circle",
    "add_circle_outline",
    "add_location",
    "add_shopping_cart",
    "add_to_photos",
    "add_to_queue",
    "adjust",
    "airline_seat_flat",
    "airline_seat_flat_angled",
    "airline_seat_individual_suite",
    "airline_seat_legroom_extra",
    "airline_seat_legroom_normal",
    "airline_seat_legroom_reduced",
    "airline_seat_recline_extra",
    "airline_seat_recline_normal",
    "airplanemode_active",
    "airplanemode_inactive",
    "airplay",
    "airport_shuttle",
    "alarm",
    "alarm_add",
    "alarm_off",
    "alarm_on",
    "album",
    "all_inclusive",
    "all_out",
    "android",
    "announcement",
    "apps",
    "archive",
    "arrow_back",
    "arrow_downward",
    "arrow_drop_down",
    "arrow_drop_down_circle",
    "arrow_drop_up",
    "arrow_forward",
    "arrow_upward",
    "art_track",
    "aspect_ratio",
    "assessment",
    "assignment",
    "assignment_ind",
    "assignment_late",
    "assignment_return",
    "assignment_returned",
    "assignment_turned_in",
    "assistant",
    "assistant_photo",
    "attach_file",
    "attach_money",
    "attachment",
    "audiotrack",
    "autorenew",
    "av_timer",
    "backspace",
    "backup",
    "battery_alert",
    "battery_charging_full",
    "battery_full",
    "battery_std",
    "battery_unknown",
    "beach_access",
    "beenhere",
    "block",
    "bluetooth",
    "bluetooth_audio",
    "bluetooth_connected",
    "bluetooth_disabled",
    "bluetooth_searching",
    "blur_circular",
    "blur_linear",
    "blur_off",
    "blur_on",
    "book",
    "bookmark",
    "bookmark_border",
    "border_all",
    "border_bottom",
    "border_clear",
    "border_color",
    "border_horizontal",
    "border_inner",
    "border_left",
    "border_outer",
    "border_right",
    "border_style",
    "border_top",
    "border_vertical",
    "branding_watermark",
    "brightness_1",
    "brightness_2",
    "brightness_3",
    "brightness_4",
    "brightness_5",
    "brightness_6",
    "brightness_7",
    "brightness_auto",
    "brightness_high",
    "brightness_low",
    "brightness_medium",
    "broken_image",
    "brush",
    "bubble_chart",
    "bug_report",
    "build",
    "burst_mode",
    "business",
    "business_center",
    "cached",
    "cake",
    "call",
    "call_end",
    "call_made",
    "call_merge",
    "call_missed",
    "call_missed_outgoing",
    "call_received",
    "call_split",
    "call_to_action",
    "camera",
    "camera_alt",
    "camera_enhance",
    "camera_front",
    "camera_rear",
    "camera_roll",
    "cancel",
    "card_giftcard",
    "card_membership",
    "card_travel",
    "casino",
    "cast",
    "cast_connected",
    "center_focus_strong",
    "center_focus_weak",
    "change_history",
    "chat",
    "chat_bubble",
    "chat_bubble_outline",
    "check",
    "check_box",
    "check_box_outline_blank",
    "check_circle",
    "chevron_left",
    "chevron_right",
    "child_care",
    "child_friendly",
    "chrome_reader_mode",
    "class",
    "clear",
    "clear_all",
    "close",
    "closed_caption",
    "cloud",
    "cloud_circle",
    "cloud_done",
    "cloud_download",
    "cloud_off",
    "cloud_queue",
    "cloud_upload",
    "code",
    "collections",
    "collections_bookmark",
    "color_lens",
    "colorize",
    "comment",
    "compare",
    "compare_arrows",
    "computer",
    "confirmation_number",
    "contact_mail",
    "contact_phone",
    "contacts",
    "content_copy",
    "content_cut",
    "content_paste",
    "control_point",
    "control_point_duplicate",
    "copyright",
    "create",
    "create_new_folder",
    "credit_card",
    "crop",
    "crop_16_9",
    "crop_3_2",
    "crop_5_4",
    "crop_7_5",
    "crop_din",
    "crop_free",
    "crop_landscape",
    "crop_original",
    "crop_portrait",
    "crop_rotate",
    "crop_square",
    "dashboard",
    "data_usage",
    "date_range",
    "dehaze",
    "delete",
    "delete_forever",
    "delete_sweep",
    "description",
    "desktop_mac",
    "desktop_windows",
    "details",
    "developer_board",
    "developer_mode",
    "device_hub",
    "devices",
    "devices_other",
    "dialer_sip",
    "dialpad",
    "directions",
    "directions_bike",
    "directions_boat",
    "directions_bus",
    "directions_car",
    "directions_railway",
    "directions_run",
    "directions_subway",
    "directions_transit",
    "directions_walk",
    "disc_full",
    "dns",
    "do_not_disturb",
    "do_not_disturb_alt",
    "do_not_disturb_off",
    "do_not_disturb_on",
    "dock",
    "domain",
    "done",
    "done_all",
    "donut_large",
    "donut_small",
    "drafts",
    "drag_handle",
    "drive_eta",
    "dvr",
    "edit",
    "edit_location",
    "eject",
    "email",
    "enhanced_encryption",
    "equalizer",
    "error",
    "error_outline",
    "euro_symbol",
    "ev_station",
    "event",
    "event_available",
    "event_busy",
    "event_note",
    "event_seat",
    "exit_to_app",
    "expand_less",
    "expand_more",
    "explicit",
    "explore",
    "exposure",
    "exposure_neg_1",
    "exposure_neg_2",
    "exposure_plus_1",
    "exposure_plus_2",
    "exposure_zero",
    "extension",
    "face",
    "fast_forward",
    "fast_rewind",
    "favorite",
    "favorite_border",
    "featured_play_list",
    "featured_video",
    "feedback",
    "fiber_dvr",
    "fiber_manual_record",
    "fiber_new",
    "fiber_pin",
    "fiber_smart_record",
    "file_download",
    "file_upload",
    "filter",
    "filter_1",
    "filter_2",
    "filter_3",
    "filter_4",
    "filter_5",
    "filter_6",
    "filter_7",
    "filter_8",
    "filter_9",
    "filter_9_plus",
    "filter_b_and_w",
    "filter_center_focus",
    "filter_drama",
    "filter_frames",
    "filter_hdr",
    "filter_list",
    "filter_none",
    "filter_tilt_shift",
    "filter_vintage",
    "find_in_page",
    "find_replace",
    "fingerprint",
    "first_page",
    "fitness_center",
    "flag",
    "flare",
    "flash_auto",
    "flash_off",
    "flash_on",
    "flight",
    "flight_land",
    "flight_takeoff",
    "flip",
    "flip_to_back",
    "flip_to_front",
    "folder",
    "folder_open",
    "folder_shared",
    "folder_special",
    "font_download",
    "format_align_center",
    "format_align_justify",
    "format_align_left",
    "format_align_right",
    "format_bold",
    "format_clear",
    "format_color_fill",
    "format_color_reset",
    "format_color_text",
    "format_indent_decrease",
    "format_indent_increase",
    "format_italic",
    "format_line_spacing",
    "format_list_bulleted",
    "format_list_numbered",
    "format_paint",
    "format_quote",
    "format_shapes",
    "format_size",
    "format_strikethrough",
    "format_textdirection_l_to_r",
    "format_textdirection_r_to_l",
    "format_underlined",
    "forum",
    "forward",
    "forward_10",
    "forward_30",
    "forward_5",
    "free_breakfast",
    "fullscreen",
    "fullscreen_exit",
    "functions",
    "g_translate",
    "gamepad",
    "games",
    "gavel",
    "gesture",
    "get_app",
    "gif",
    "golf_course",
    "gps_fixed",
    "gps_not_fixed",
    "gps_off",
    "grade",
    "gradient",
    "grain",
    "graphic_eq",
    "grid_off",
    "grid_on",
    "group",
    "group_add",
    "group_work",
    "hd",
    "hdr_off",
    "hdr_on",
    "hdr_strong",
    "hdr_weak",
    "headset",
    "headset_mic",
    "healing",
    "hearing",
    "help",
    "help_outline",
    "high_quality",
    "highlight",
    "highlight_off",
    "history",
    "home",
    "hot_tub",
    "hotel",
    "hourglass_empty",
    "hourglass_full",
    "http",
    "https",
    "image",
    "image_aspect_ratio",
    "import_contacts",
    "import_export",
    "important_devices",
    "inbox",
    "indeterminate_check_box",
    "info",
    "info_outline",
    "input",
    "insert_chart",
    "insert_comment",
    "insert_drive_file",
    "insert_emoticon",
    "insert_invitation",
    "insert_link",
    "insert_photo",
    "invert_colors",
    "invert_colors_off",
    "iso",
    "keyboard",
    "keyboard_arrow_down",
    "keyboard_arrow_left",
    "keyboard_arrow_right",
    "keyboard_arrow_up",
    "keyboard_backspace",
    "keyboard_capslock",
    "keyboard_hide",
    "keyboard_return",
    "keyboard_tab",
    "keyboard_voice",
    "kitchen",
    "label",
    "label_outline",
    "landscape",
    "language",
    "laptop",
    "laptop_chromebook",
    "laptop_mac",
    "laptop_windows",
    "last_page",
    "launch",
    "layers",
    "layers_clear",
    "leak_add",
    "leak_remove",
    "lens",
    "library_add",
    "library_books",
    "library_music",
    "lightbulb_outline",
    "line_style",
    "line_weight",
    "linear_scale",
    "link",
    "linked_camera",
    "list",
    "live_help",
    "live_tv",
    "local_activity",
    "local_airport",
    "local_atm",
    "local_bar",
    "local_cafe",
    "local_car_wash",
    "local_convenience_store",
    "local_dining",
    "local_drink",
    "local_florist",
    "local_gas_station",
    "local_grocery_store",
    "local_hospital",
    "local_hotel",
    "local_laundry_service",
    "local_library",
    "local_mall",
    "local_movies",
    "local_offer",
    "local_parking",
    "local_pharmacy",
    "local_phone",
    "local_pizza",
    "local_play",
    "local_post_office",
    "local_printshop",
    "local_see",
    "local_shipping",
    "local_taxi",
    "location_city",
    "location_disabled",
    "location_off",
    "location_on",
    "location_searching",
    "lock",
    "lock_open",
    "lock_outline",
    "looks",
    "looks_3",
    "looks_4",
    "looks_5",
    "looks_6",
    "looks_one",
    "looks_two",
    "loop",
    "loupe",
    "low_priority",
    "loyalty",
    "mail",
    "mail_outline",
    "map",
    "markunread",
    "markunread_mailbox",
    "memory",
    "menu",
    "merge_type",
    "message",
    "mic",
    "mic_none",
    "mic_off",
    "mms",
    "mode_comment",
    "mode_edit",
    "monetization_on",
    "money_off",
    "monochrome_photos",
    "mood",
    "mood_bad",
    "more",
    "more_horiz",
    "more_vert",
    "motorcycle",
    "mouse",
    "move_to_inbox",
    "movie",
    "movie_creation",
    "movie_filter",
    "multiline_chart",
    "music_note",
    "music_video",
    "my_location",
    "nature",
    "nature_people",
    "navigate_before",
    "navigate_next",
    "navigation",
    "near_me",
    "network_cell",
    "network_check",
    "network_locked",
    "network_wifi",
    "new_releases",
    "next_week",
    "nfc",
    "no_encryption",
    "no_sim",
    "not_interested",
    "note",
    "note_add",
    "notifications",
    "notifications_active",
    "notifications_none",
    "notifications_off",
    "notifications_paused",
    "offline_pin",
    "ondemand_video",
    "opacity",
    "open_in_browser",
    "open_in_new",
    "open_with",
    "pages",
    "pageview",
    "palette",
    "pan_tool",
    "panorama",
    "panorama_fish_eye",
    "panorama_horizontal",
    "panorama_vertical",
    "panorama_wide_angle",
    "party_mode",
    "pause",
    "pause_circle_filled",
    "pause_circle_outline",
    "payment",
    "people",
    "people_outline",
    "perm_camera_mic",
    "perm_contact_calendar",
    "perm_data_setting",
    "perm_device_information",
    "perm_identity",
    "perm_media",
    "perm_phone_msg",
    "perm_scan_wifi",
    "person",
    "person_add",
    "person_outline",
    "person_pin",
    "person_pin_circle",
    "personal_video",
    "pets",
    "phone",
    "phone_android",
    "phone_bluetooth_speaker",
    "phone_forwarded",
    "phone_in_talk",
    "phone_iphone",
    "phone_locked",
    "phone_missed",
    "phone_paused",
    "phonelink",
    "phonelink_erase",
    "phonelink_lock",
    "phonelink_off",
    "phonelink_ring",
    "phonelink_setup",
    "photo",
    "photo_album",
    "photo_camera",
    "photo_filter",
    "photo_library",
    "photo_size_select_actual",
    "photo_size_select_large",
    "photo_size_select_small",
    "picture_as_pdf",
    "picture_in_picture",
    "picture_in_picture_alt",
    "pie_chart",
    "pie_chart_outlined",
    "pin_drop",
    "place",
    "play_arrow",
    "play_circle_filled",
    "play_circle_outline",
    "play_for_work",
    "playlist_add",
    "playlist_add_check",
    "playlist_play",
    "plus_one",
    "poll",
    "polymer",
    "pool",
    "portable_wifi_off",
    "portrait",
    "power",
    "power_input",
    "power_settings_new",
    "pregnant_woman",
    "present_to_all",
    "print",
    "priority_high",
    "public",
    "publish",
    "query_builder",
    "question_answer",
    "queue",
    "queue_music",
    "queue_play_next",
    "radio",
    "radio_button_checked",
    "radio_button_unchecked",
    "rate_review",
    "receipt",
    "recent_actors",
    "record_voice_over",
    "redeem",
    "redo",
    "refresh",
    "remove",
    "remove_circle",
    "remove_circle_outline",
    "remove_from_queue",
    "remove_red_eye",
    "remove_shopping_cart",
    "reorder",
    "repeat",
    "repeat_one",
    "replay",
    "replay_10",
    "replay_30",
    "replay_5",
    "reply",
    "reply_all",
    "report",
    "report_problem",
    "restaurant",
    "restaurant_menu",
    "restore",
    "restore_page",
    "ring_volume",
    "room",
    "room_service",
    "rotate_90_degrees_ccw",
    "rotate_left",
    "rotate_right",
    "rounded_corner",
    "router",
    "rowing",
    "rss_feed",
    "rv_hookup",
    "satellite",
    "save",
    "scanner",
    "schedule",
    "school",
    "screen_lock_landscape",
    "screen_lock_portrait",
    "screen_lock_rotation",
    "screen_rotation",
    "screen_share",
    "sd_card",
    "sd_storage",
    "search",
    "security",
    "select_all",
    "send",
    "sentiment_dissatisfied",
    "sentiment_neutral",
    "sentiment_satisfied",
    "sentiment_very_dissatisfied",
    "sentiment_very_satisfied",
    "settings",
    "settings_applications",
    "settings_backup_restore",
    "settings_bluetooth",
    "settings_brightness",
    "settings_cell",
    "settings_ethernet",
    "settings_input_antenna",
    "settings_input_component",
    "settings_input_composite",
    "settings_input_hdmi",
    "settings_input_svideo",
    "settings_overscan",
    "settings_phone",
    "settings_power",
    "settings_remote",
    "settings_system_daydream",
    "settings_voice",
    "share",
    "shop",
    "shop_two",
    "shopping_basket",
    "shopping_cart",
    "short_text",
    "show_chart",
    "shuffle",
    "signal_cellular_4_bar",
    "signal_cellular_connected_no_internet_4_bar",
    "signal_cellular_no_sim",
    "signal_cellular_null",
    "signal_cellular_off",
    "signal_wifi_4_bar",
    "signal_wifi_4_bar_lock",
    "signal_wifi_off",
    "sim_card",
    "sim_card_alert",
    "skip_next",
    "skip_previous",
    "slideshow",
    "slow_motion_video",
    "smartphone",
    "smoke_free",
    "smoking_rooms",
    "sms",
    "sms_failed",
    "snooze",
    "sort",
    "sort_by_alpha",
    "spa",
    "space_bar",
    "speaker",
    "speaker_group",
    "speaker_notes",
    "speaker_notes_off",
    "speaker_phone",
    "spellcheck",
    "star",
    "star_border",
    "star_half",
    "stars",
    "stay_current_landscape",
    "stay_current_portrait",
    "stay_primary_landscape",
    "stay_primary_portrait",
    "stop",
    "stop_screen_share",
    "storage",
    "store",
    "store_mall_directory",
    "straighten",
    "streetview",
    "strikethrough_s",
    "style",
    "subdirectory_arrow_left",
    "subdirectory_arrow_right",
    "subject",
    "subscriptions",
    "subtitles",
    "subway",
    "supervisor_account",
    "surround_sound",
    "swap_calls",
    "swap_horiz",
    "swap_vert",
    "swap_vertical_circle",
    "switch_camera",
    "switch_video",
    "sync",
    "sync_disabled",
    "sync_problem",
    "system_update",
    "system_update_alt",
    "tab",
    "tab_unselected",
    "tablet",
    "tablet_android",
    "tablet_mac",
    "tag_faces",
    "tap_and_play",
    "terrain",
    "text_fields",
    "text_format",
    "textsms",
    "texture",
    "theaters",
    "thumb_down",
    "thumb_up",
    "thumbs_up_down",
    "time_to_leave",
    "timelapse",
    "timeline",
    "timer",
    "timer_10",
    "timer_3",
    "timer_off",
    "title",
    "toc",
    "today",
    "toll",
    "tonality",
    "touch_app",
    "toys",
    "track_changes",
    "traffic",
    "train",
    "tram",
    "transfer_within_a_station",
    "transform",
    "translate",
    "trending_down",
    "trending_flat",
    "trending_up",
    "tune",
    "turned_in",
    "turned_in_not",
    "tv",
    "unarchive",
    "undo",
    "unfold_less",
    "unfold_more",
    "update",
    "usb",
    "verified_user",
    "vertical_align_bottom",
    "vertical_align_center",
    "vertical_align_top",
    "vibration",
    "video_call",
    "video_label",
    "video_library",
    "videocam",
    "videocam_off",
    "videogame_asset",
    "view_agenda",
    "view_array",
    "view_carousel",
    "view_column",
    "view_comfy",
    "view_compact",
    "view_day",
    "view_headline",
    "view_list",
    "view_module",
    "view_quilt",
    "view_stream",
    "view_week",
    "vignette",
    "visibility",
    "visibility_off",
    "voice_chat",
    "voicemail",
    "volume_down",
    "volume_mute",
    "volume_off",
    "volume_up",
    "vpn_key",
    "vpn_lock",
    "wallpaper",
    "warning",
    "watch",
    "watch_later",
    "wb_auto",
    "wb_cloudy",
    "wb_incandescent",
    "wb_iridescent",
    "wb_sunny",
    "wc",
    "web",
    "web_asset",
    "weekend",
    "whatshot",
    "widgets",
    "wifi",
    "wifi_lock",
    "wifi_tethering",
    "work",
    "wrap_text",
    "youtube_searched_for",
    "zoom_in",
    "zoom_out",
    "zoom_out_map",
]



"""Implementation of the PerspectiveViewer Web Component"""

import panel as pn
import param

from awesome_panel_extensions.bokeh_extensions.pivot_table import PivotTable as _BkPivotTable
from awesome_panel_extensions.widgets.dataframe_base import DataFrameWithStreamAndPatchBaseWidget

# This is need to be able to use Perspective in notebook via pn.extension("perspective")
pn.extension._imports[  # pylint: disable=protected-access
    "pivottable"
] = "awesome_panel_extensions.bokeh_extensions.pivot_table"


class PivotTable(DataFrameWithStreamAndPatchBaseWidget):  # pylint: disable=abstract-method
    """The PivotTable widget enables exploring large tables of data
    in an interactive pivot table"""

    _widget_type = _BkPivotTable

    # I set this to something > 0. Otherwise the PerspectiveViewer widget will have a height of 0px
    # It will appear as if it does not work.
    height = param.Integer(default=300, bounds=(0, None))



"""Implementation of the Predix App Navigation Widget

See https://www.predix-ui.com/#/elements/px-app-nav"""

from bokeh.core import properties
from bokeh.models import ColumnDataSource
from bokeh.models.layouts import HTMLBox


class WebComponent(HTMLBox):
    """A Bokeh Model that enables easily creating new Panel components from web components

    See https://www.predix-ui.com/#/elements/px-app-nav"""

    componentType = properties.String("htmlbox")
    innerHTML = properties.String("")
    attributesToWatch = properties.Dict(properties.String, properties.Any)
    attributesLastChange = properties.Dict(properties.String, properties.Any)
    propertiesToWatch = properties.Dict(properties.String, properties.Any)
    propertiesLastChange = properties.Dict(properties.String, properties.Any)
    eventsToWatch = properties.Dict(properties.String, properties.Any)
    eventsCountLastChange = properties.Dict(properties.String, properties.Int)
    columnDataSource = properties.Instance(ColumnDataSource)
    columnDataSourceOrient = properties.String()
    columnDataSourceLoadFunction = properties.String()



"""Exports"""
from awesome_panel_extensions.developer_tools.designer.designer import Designer
from awesome_panel_extensions.developer_tools.designer.services.component_reloader import (
    ComponentReloader,
)



"""This module contains the Material LinearProgress and CircularProgress"""
import param

from awesome_panel_extensions.frameworks._base.progress import Progress as _BaseProgress

_ATTRIBUTES_TO_WATCH_BASE = {"class": "bar_color"}
_PROPERTIES_TO_WATCH_BASE = {
    "indeterminate": "active",
    "progress": "_progress",
    "closed": "closed",
}

DENSITY_RATIO = 4


class _MaterialProgress(_BaseProgress):
    closed = param.Boolean(
        default=False,
        doc="""Sets the progress indicator to the closed state. Sets content opacity to 0.
        Typically should be set to true when loading has finished.""",
    )
    _progress = param.Number(default=None, bounds=(0, 1), allow_None=True)

    attributes_to_watch = param.Dict(_ATTRIBUTES_TO_WATCH_BASE)
    properties_to_watch = param.Dict(_PROPERTIES_TO_WATCH_BASE)

    def __init__(self, **params):
        # Hack: to make sure that value is shown on construction
        if "value" in params and "active" not in params:
            params["active"] = False

        super().__init__(**params)

        self._update_progress()

    @param.depends("value", "max", watch=True)
    def _update_progress(self, *_):
        if self.value is None or self.value == 0 or self.max is None or self.max == 0:
            self._progress = None
        else:
            self._progress = self.value / self.max


class LinearProgress(_MaterialProgress):
    """The Material LinearProgress widget conveys progress information to the user"""

    html = param.String(
        "<mwc-linear-progress style='width:100%' progress='0.0'></mwc-linear-progress"
    )

    buffer = param.Integer(
        default=None,
        bounds=(0, 100),
        allow_None=True,
        doc="""Sets the buffer progress bar's value. Value should be between [0, 1].
            Setting this value to be less than max will reveal moving, buffering dots.""",
    )

    _buffer = param.Number(default=None, allow_None=True,)
    reverse = param.Boolean(
        default=False, doc="Reverses the direction of the linear progress indicator."
    )

    properties_to_watch = param.Dict(
        {**_PROPERTIES_TO_WATCH_BASE, "buffer": "_buffer", "reverse": "reverse"}
    )

    @param.depends("max", watch=True)
    def _update_buffer_bounds(self):
        self.param.buffer.bounds = (0, self.max)

    @param.depends("buffer", "max", watch=True)
    def _update_buffer(self, *_):
        if self.buffer is None or self.buffer == 0 or self.max is None or self.max == 0:
            self._buffer = None
        else:
            self._buffer = self.buffer / self.max


class CircularProgress(_MaterialProgress):
    """The Material Circular Progress Widget conveys progress information to the user"""

    html = param.String("<mwc-circular-progress style='width:51px'></mwc-circular-progress")

    density = param.Integer(
        default=0,
        bounds=(-8, 2000),
        doc="""Sets the progress indicator's sizing based on density scale. Minimum value is -8.
        Each unit change in density scale corresponds to 4px change in side dimensions. The stroke
        width adjusts automatically.""",
    )
    # _style = param.String()

    # attributes_to_watch = param.Dict(dict(**_ATTRIBUTES_TO_WATCH_BASE, style="_style"))
    properties_to_watch = param.Dict(dict(**_PROPERTIES_TO_WATCH_BASE, density="density"))

    def __init__(self, **params):
        if "density" in params and "html" not in params:
            density = params["density"]
            diameter = round((density + 8) * DENSITY_RATIO + 17)
            html = (
                f"<mwc-circular-progress style='height:{diameter}px;width:{diameter}px' "
                f"density={density}></mwc-circular-progress"
            )
            params["html"] = html

        super().__init__(**params)

        self._update_diameter()

    @param.depends("density", watch=True)
    def _update_diameter(self, *_):
        diameter = round((self.density + 8) * DENSITY_RATIO + 17)
        self.min_height = diameter
        self.min_width = diameter
        # Cannot get the style updating programmatically. Starts an infinite loop.
        # self._style = f"height:{diameter}px;width:{diameter}px;"



"""This module contains base functions and classes used to implement
'DataFrame' widgets like the PerspectiveViewer and Tabulator"""
from typing import Any, Dict, Union

import pandas as pd
import panel as pn
import param
from bokeh.models import ColumnDataSource
from panel.widgets.base import Widget


class DataFrameWithStreamAndPatchBaseWidget(Widget):
    """Base Class for the DataFrame based widgets like PerspectiveViewer, Tabulator etc widgets"""

    value = param.DataFrame(
        doc="""A pandas.DataFrame

        Please note when specifying a Pandas.Dataframe we currently have some narrow requirements
        """,
    )
    _source = param.ClassSelector(
        class_=ColumnDataSource,
        doc="""Used to transfer the `value` efficiently to frontend. \
        Should always be in sync with value""",
    )
    # pylint: disable=line-too-long
    # Need this because ColumnDataSource not provides streamed value only streamed events
    # https://discourse.bokeh.org/t/how-do-i-get-the-new-data-only-when-i-stream-via-a-columndatasource/6186/2
    # pylint: enable=line-too-long
    _source_stream = param.ClassSelector(
        class_=ColumnDataSource,
        doc="""Used to transfer a streamed (i.e appended) `value` efficiently to frontend.""",
    )
    # pylint: disable=line-too-long
    # Need this because ColumnDataSource not provides patched value only patched events
    # https://discourse.bokeh.org/t/how-do-i-get-the-new-data-only-when-i-stream-via-a-columndatasource/6186/2
    # pylint: enable=line-too-long
    _source_patch = param.ClassSelector(
        class_=ColumnDataSource,
        doc="""Used to transfer a patched (i.e updated) `value` efficiently to frontend.""",
    )
    _rename = {
        "value": None,
        "_source": "source",
        "_source_stream": "source_stream",
        "_source_patch": "source_patch",
    }

    def __init__(self, **params):
        super().__init__(**params)

        self._pause_cds_updates = False
        self._source_patch = ColumnDataSource({})
        self._source_stream = ColumnDataSource({})
        self._set_source()

    @param.depends("value", watch=True)
    def _set_source(self):
        if self._pause_cds_updates:
            return

        if isinstance(self.value, pd.DataFrame) and (
            not isinstance(self.value.index, pd.RangeIndex)
            or self.value.index.start != 0
            or self.value.index.step != 1
        ):
            raise ValueError(
                "Please provide a DataFrame with RangeIndex starting at 0 and with step 1"
            )

        if self._source is None:
            if self.value is None:
                self._source = ColumnDataSource({})
            else:
                self._source = ColumnDataSource(self.value)
        else:
            if self.value is None:
                self._source.data = {}
            else:
                self._source.data = self.value

    def stream(self, stream_value: Union[pd.DataFrame, pd.Series, Dict], reset_index: bool = True):
        """Streams (appends) the `stream_value` provided to the existing value in an efficient
        manner.

        Args:
            stream_value (Union[pd.DataFrame, pd.Series, Dict]): The new value(s) to append to the
                existing value.
            reset_index (bool, optional): If the stream_value is a DataFrame and `reset_index` is
                True then the index of it is reset if True. Helps to keep the index unique and
                named `index`. Defaults to True.

        Raises:
            ValueError: Raised if the stream_value is not a supported type.

        Example: Stream a Series to a DataFrame

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> stream_value = pd.Series({"x": 4, "y": "d"})
        >>> widget.stream(stream_value)
        >>> widget.value.to_dict("list")
        {'x': [1, 2, 4], 'y': ['a', 'b', 'd']}

        Example: Stream a Dataframe to a Dataframe

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> stream_value = pd.DataFrame({"x": [3, 4], "y": ["c", "d"]})
        >>> widget.stream(stream_value)
        >>> widget.value.to_dict("list")
        {'x': [1, 2, 3, 4], 'y': ['a', 'b', 'c', 'd']}


        Example: Stream a Dictionary row to a DataFrame

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> stream_value = {"x": 4, "y": "d"}
        >>> widget.stream(stream_value)
        >>> widget.value.to_dict("list")
        {'x': [1, 2, 4], 'y': ['a', 'b', 'd']}

        Example: Stream a Dictionary of Columns to a Dataframe

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> stream_value = {"x": [3, 4], "y": ["c", "d"]}
        >>> widget.stream(stream_value)
        >>> widget.value.to_dict("list")
        {'x': [1, 2, 3, 4], 'y': ['a', 'b', 'c', 'd']}
        """

        if isinstance(self.value, pd.DataFrame):
            if isinstance(stream_value, pd.DataFrame):
                self._stream_dataframe(stream_value, reset_index)
            elif isinstance(stream_value, pd.Series):
                self._stream_series(stream_value)
            elif isinstance(stream_value, dict):
                if stream_value:
                    try:
                        stream_value = pd.DataFrame(stream_value)
                    except ValueError:
                        stream_value = pd.Series(stream_value)
                    self.stream(stream_value)
            else:
                raise ValueError("The patch value provided is not a DataFrame, Series or Dict!")
        else:
            self._pause_cds_updates = True
            self._source.stream(stream_value)
            self._source_stream.data = stream_value
            self.param.trigger("value")
            self._pause_cds_updates = False

    def _stream_dataframe(self, stream_value, reset_index):
        value_index_start = self.value.index.max() + 1
        if reset_index:
            stream_value = stream_value.reset_index(drop=True)
            stream_value.index += value_index_start
        self._pause_cds_updates = True
        self._source.stream(stream_value)
        new_value = pd.concat([self.value, stream_value])
        value_index_end = new_value.index.max() + 1
        stream_value2 = (
            new_value.loc[value_index_start:value_index_end,].reset_index().to_dict("list")
        )
        self._source_stream.data = stream_value2
        self.value = new_value
        self._pause_cds_updates = False

    def _stream_series(self, stream_value):
        value_index_start = self.value.index.max() + 1
        self._pause_cds_updates = True
        self.value.loc[value_index_start] = stream_value
        self._source.stream(stream_value)
        stream_value2 = {k: [v] for k, v in stream_value.to_dict().items()}
        stream_value2["index"] = [value_index_start]
        self._source_stream.data = stream_value2
        self.param.trigger("value")
        self._pause_cds_updates = False

    def patch(self, patch_value: Union[pd.DataFrame, pd.Series, Dict]):
        """Patches (updates) the existing value with the `patch_value` in an efficient manner.

        Args:
            patch_value (Union[pd.DataFrame, pd.Series, Dict]): The value(s) to patch the
                existing value with.

        Raises:
            ValueError: Raised if the patch_value is not a supported type.



        Example: Patch a DataFrame with a Dictionary row.

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> patch_value = {"x": [(0, 3)]}
        >>> widget.patch(patch_value)
        >>> widget.value.to_dict("list")
        {'x': [3, 2], 'y': ['a', 'b']}

        Example: Patch a Dataframe with a Dictionary of Columns.

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> patch_value = {"x": [(slice(2), (3,4))], "y": [(1,'d')]}
        >>> widget.patch(patch_value)
        >>> widget.value.to_dict("list")
        {'x': [3, 4], 'y': ['a', 'd']}

        Example: Patch a DataFrame with a Series. Please note the index is used in the update

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> patch_value = pd.Series({"index": 1, "x": 4, "y": "d"})
        >>> widget.patch(patch_value)
        >>> widget.value.to_dict("list")
        {'x': [1, 4], 'y': ['a', 'd']}

        Example: Patch a Dataframe with a Dataframe. Please note the index is used in the update.

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> widget = DataFrameWithStreamAndPatchBaseWidget(value=value)
        >>> patch_value = pd.DataFrame({"x": [3, 4], "y": ["c", "d"]})
        >>> widget.patch(patch_value)
        >>> widget.value.to_dict("list")
        {'x': [3, 4], 'y': ['c', 'd']}
        """
        if isinstance(self.value, pd.DataFrame):
            if isinstance(patch_value, pd.DataFrame):
                patch_value_dict = self._patch_dataframe_to_dict(patch_value)
                self.patch(patch_value_dict)
            elif isinstance(patch_value, pd.Series):
                patch_value_dict = self._patch_series_to_dict(patch_value)
                self.patch(patch_value_dict)
            elif isinstance(patch_value, dict):
                self._patch_from_dict(patch_value)

            else:
                raise ValueError(
                    f"""Patching a patch_value of type {type(patch_value)} is not supported.
                    Please provide a DataFrame, Series or Dict"""
                )
        else:
            if isinstance(patch_value, dict):
                self._pause_cds_updates = True
                self._cds_patch(patch_value)
                self.param.trigger("value")
                self._pause_cds_updates = False
            else:
                raise ValueError(
                    f"""Patching a patch_value of type {type(patch_value)} is not supported.
                    Please provide a dict"""
                )

    @staticmethod
    def _patch_dataframe_to_dict(patch_value: pd.DataFrame) -> Dict:
        patch_value_dict: Dict[Any, Any] = {}
        for column in patch_value.columns:
            patch_value_dict[column] = []
            for index in patch_value.index:
                patch_value_dict[column].append((index, patch_value.loc[index, column]))

        return patch_value_dict

    @staticmethod
    def _patch_series_to_dict(patch_value: pd.Series) -> Dict:
        if "index" in patch_value:  # Series orient is row
            patch_value_dict = {k: [(patch_value["index"], v)] for k, v in patch_value.items()}
            patch_value_dict.pop("index")
        else:  # Series orient is column
            patch_value_dict = {patch_value.name: list(patch_value.items())}
        return patch_value_dict

    def _patch_from_dict(self, patch_value: Dict):
        self._pause_cds_updates = True
        for key, value in patch_value.items():
            for update in value:
                self.value.loc[update[0], key] = update[1]
        self._cds_patch(patch_value)
        self.param.trigger("value")
        self._pause_cds_updates = False

    def _cds_patch(self, patch_value):
        self._source.patch(patch_value)
        self._source_patch.data = patch_value

    @classmethod
    def config(cls):
        """Adds the required js files to pn.config.js_files"""
        for key, value in cls._widget_type.__js_require__["paths"].items():
            pn.config.js_files[key.replace("-", "")] = value + ".js"



"""The PanelLinkButton displayes the Panel Logo and if clicked opens the Panel site"""
import panel as pn
import param

from awesome_panel_extensions.widgets.link_buttons.image_link_button import (
    _STYLE,
    DerivedImageLinkButton,
)

THEMES = ["light", "dark"]
IMAGE_URLS = {
    "light": "https://panel.holoviz.org/_static/logo_stacked.png",
    "dark": "https://panel.holoviz.org/_static/logo_horizontal.png",
}
_DARK_STYLE = {
    "background": "black",
    "cursor": "pointer",
    "border": "1px solid #ddd",
    "border-radius": "4px",
    "padding": "5px",
}
LAYOUTS = {
    "light": {"height": 31, "width": 37, "background": "white", "style": _STYLE},
    "dark": {"height": 31, "width": 140, "background": "black", "style": _DARK_STYLE},
}
LINK_URL = "https://panel.holoviz.org"


class PanelLinkButton(DerivedImageLinkButton):
    """The PanelLinkButton displayes the Panel Logo and if clicked opens the Panel site"""

    theme = param.ObjectSelector(default="light", objects=THEMES)

    image_url = param.String(doc="The url to the image", constant=True)
    link_url = param.String(default=LINK_URL, doc="The url to open when clicked", constant=True)

    def __init__(self, **params):
        super().__init__(**params)

        self._update_image_url_from_theme()

    @param.depends("theme", watch=True)
    def _update_image_url_from_theme(self, *events):  # pylint: disable=unused-argument
        with param.edit_constant(self):
            self.image_url = IMAGE_URLS[self.theme]
            height = LAYOUTS[self.theme]["height"]
            width = LAYOUTS[self.theme]["width"]
            if self.height:
                self.width = int(self.height * width / height)
            else:
                self.height = height
                self.width = width
            self.style = LAYOUTS[self.theme]["style"]


if __name__.startswith("bokeh"):
    button = PanelLinkButton()
    settings_pane = pn.Param(
        button,
        parameters=["theme", "height", "width", "sizing_mode", "margin"],
        background="lightgray",
        sizing_mode="stretch_width",
    )
    app = pn.Column(button, settings_pane, width=500, height=800)
    app.servable()



"""This module implements the StoppedComponent. The view of the StoppedComponent is shown to the
user when he/ she exits the Designer"""
import panel as pn
import param

from awesome_panel_extensions import assets

MESSAGE = f"""\
# Stopped

Thanks for using the Awesome Panel Designer by awesome-panel.org

<br>

![Logo]({assets.SPINNER_PANEL_STATIC_LIGHT_400_340})

"""


class StoppedComponent(param.Parameterized):
    """The viewStoppedComponent is shown to the user when he/ she exits the Designer"""

    view = param.ClassSelector(class_=pn.pane.Markdown)

    def __init__(self, **params):
        super().__init__(**params)

        self.view = pn.pane.Markdown(MESSAGE)



"""The Tabulator Pane wraps the [Tabulator](http://tabulator.info/) table to provide an awesome
interative table.

You can
- Specify an initial `configuration`.
- Provide a `value` as a Pandas DataFrame or Bokeh ColumnDataSource.
- Edit cell values in the browser.
- Select rows in the browser or in code.
- `stream` (append) to the `value`.
- `patch` (update) the `value`.
- Change the (css) style using the TabulatorStylesheet.
"""
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import panel as pn
import param
from bokeh.models.sources import ColumnDataSource
from panel.widgets.base import Widget

from awesome_panel_extensions.bokeh_extensions.tabulator_model import CSS_HREFS
from awesome_panel_extensions.bokeh_extensions.tabulator_model import TabulatorModel as _BkTabulator

_DEFAULT_CONFIGURATION = {"autoColumns": True}
_FORMATTERS = {
    "bool": "plaintext",
    "category": "plaintext",
    "datetime64": "datetime",
    "datetime64[ns]": "datetime",
    "float64": "money",
    "int64": "money",
    "object": "plaintext",
    "timedelta[ns]": "datetimediff",
}
_SORTERS = {
    "bool": "tickCross",
    "category": "string",
    "datetime64": "datetime",
    "datetime64[ns]": "datetime",
    "float64": "number",
    "int64": "number",
    "object": "string",
    "timedelta[ns]": "datetime",
}
_HOZ_ALIGNS = {
    "bool": "center",
    "category": "left",
    "datetime64": "left",
    "datetime64[ns]": "left",
    "float64": "right",
    "int64": "right",
    "object": "left",
    "timedelta[ns]": "right",
}


class Tabulator(Widget):
    """The Tabulator Pane wraps the [Tabulator](http://tabulator.info/) table to provide an
awesome interative table.

You can
- Specify a `configuration` dictionary at instantation. See http://tabulator.info/.
- Provide an initial `value` as a Pandas DataFrame or Bokeh ColumnDataSource.
- `stream` (append) to the `value`.
- `patch` (update) the `value`.

Example: Data specified in configuration

>>> from awesome_panel_extensions.widgets.tabulator import Tabulator
>>> configuration = {
...     "layout": "fitColumns",
...     "data": [
...         {"x": [1], "y": 'a'},
...         {"x": [2], "y": 'b'}
...         ],
...     "initialSort":[
...         {"column":"y", "dir":"desc"},
...     ],
...     "columns":[
...         {"title": "Value", "field":"x"},
...         {"title": "Item", "field":"y", "hozAlign":"right", "formatter":"money"}
...     ],
... }
>>> Tabulator(configuration=configuration)
Tabulator(...)

Example: Data specified as Pandas.DataFrame value

>>> import pandas as pd
>>> configuration = {
...     "layout": "fitColumns",
...     "initialSort":[
...         {"column":"y", "dir":"desc"},
...     ],
...     "columns":[
...         {"title": "Value", "field":"x"},
...         {"title": "Item", "field":"y", "hozAlign":"right", "formatter":"money"}
...     ],
... }
>>> value = pd.DataFrame([
...     {"x": [1], "y": 'a'},
...     {"x": [2], "y": 'b'}
... ])
>>> Tabulator(configuration=configuration, value=value)
Tabulator(...)

Example: Data specified as Bokeh ColumnDataSource value

>>> configuration = {
...     "layout": "fitColumns",
...     "initialSort":[
...         {"column":"y", "dir":"desc"},
...     ],
...     "columns":[
...         {"title": "Value", "field":"x"},
...         {"title": "Item", "field":"y", "hozAlign":"right", "formatter":"money"}
...     ],
... }
>>> value = ColumnDataSource({"x": [1,2], "y": ["a", "b"]})
>>> Tabulator(configuration=configuration, value=value)
Tabulator(...)
"""

    value = param.Parameter(
        doc="""One of pandas.DataFrame or bokeh.models.ColumnDataSource.

        Please note when specifying a Pandas.Dataframe we currently have some narrow requirements

        - The index should be a single index which should be called 'index' and be unique. To make
        sure things work we suggest you `.reset_index()` before usage.
        """
    )
    selection = param.List(doc="The list of selected row indexes. For example [1,4]. Default is []")
    configuration = param.Dict(
        constant=True,
        doc="""The initial Tabulator configuration. See https://tabulator.info for lots of
        examples.

        If None is provided at instantiation, then {'autocolumns': True} is added

        Please note that in order to get things working we add the below to the configuration
        on the js side.
        {
            "rowSelectionChanged": rowSelectionChanged,
            "cellEdited": cellEdited,
            "index": "index",
        }
        """,
    )

    height = param.Integer(
        default=300,
        bounds=(0, None),
        doc="""The height of the Tabulator table. Specifying a height is mandatory.""",
    )

    _source = param.ClassSelector(
        class_=ColumnDataSource, doc="Used to transfer the `value` efficiently to frontend"
    )
    _cell_change = param.Dict(
        doc="""Changed whenever the user updates a cell in the client. Sends something like
        {"c": "<COLUMN NAME>", "i": "<INDEX>", "v": "<NEW VALUE>"}. Used to transfer the change
        efficiently and update the DataFrame as I could not find a similar property/ event on the
        ColumnDataSource"""
    )

    _rename = {
        "value": None,
        "selection": None,
        "_source": "source",
    }
    _widget_type = _BkTabulator

    def __init__(self, **params):
        """The Tabulator Pane wraps the [Tabulator](http://tabulator.info/) table to provide an
        awesome interative table.

You can
- Specify a `configuration` dictionary at instantation. See http://tabulator.info/.
- Provide an initial `value` as a Pandas DataFrame or Bokeh ColumnDataSource.
- `stream` (append) to the `value`.
- `patch` (update) the `value`.

Example: Data specified in configuration

>>> from awesome_panel_extensions.widgets.tabulator import Tabulator
>>> configuration = {
...     "layout": "fitColumns",
...     "data": [
...         {"x": [1], "y": 'a'},
...         {"x": [2], "y": 'b'}
...         ],
...     "initialSort":[
...         {"column":"y", "dir":"desc"},
...     ],
...     "columns":[
...         {"title": "Value", "field":"x"},
...         {"title": "Item", "field":"y", "hozAlign":"right", "formatter":"money"}
...     ],
... }
>>> Tabulator(configuration=configuration)
Tabulator(...)

Example: Data specified as Pandas.DataFrame value

>>> import pandas as pd
>>> configuration = {
...     "layout": "fitColumns",
...     "initialSort":[
...         {"column":"y", "dir":"desc"},
...     ],
...     "columns":[
...         {"title": "Value", "field":"x"},
...         {"title": "Item", "field":"y", "hozAlign":"right", "formatter":"money"}
...     ],
... }
>>> value = pd.DataFrame([
...     {"x": [1], "y": 'a'},
...     {"x": [2], "y": 'b'}
... ])
>>> Tabulator(configuration=configuration, value=value)
Tabulator(...)

Example: Data specified as Bokeh ColumnDataSource value

>>> configuration = {
...     "layout": "fitColumns",
...     "initialSort":[
...         {"column":"y", "dir":"desc"},
...     ],
...     "columns":[
...         {"title": "Value", "field":"x"},
...         {"title": "Item", "field":"y", "hozAlign":"right", "formatter":"money"}
...     ],
... }
>>> value = ColumnDataSource({"x": [1,2], "y": ["a", "b"]})
>>> Tabulator(configuration=configuration, value=value)
Tabulator(...)
"""
        if "configuration" not in params:
            params["configuration"] = _DEFAULT_CONFIGURATION.copy()
        if "selection" not in params:
            params["selection"] = []

        super().__init__(**params)

        self._pause_cds_updates = False
        self._update_column_data_source()

    @param.depends("value", watch=True)
    def _update_column_data_source(self, *_):
        if self._pause_cds_updates:
            return

        if self.value is None:
            self._source = ColumnDataSource({})
        elif isinstance(self.value, pd.DataFrame):
            if (
                not isinstance(self.value.index, pd.RangeIndex)
                or self.value.index.start != 0
                or self.value.index.step != 1
            ):
                raise ValueError(
                    "Please provide a DataFrame with RangeIndex starting at 0 and with step 1"
                )

            if self._source:
                self._source.data = self.value
            else:
                self._source = ColumnDataSource(self.value)
        elif isinstance(self.value, ColumnDataSource):
            self._source = self.value
        else:
            raise ValueError("The `data` provided is not of a supported type!")

    @param.depends("_cell_change", watch=True)
    def _update_value_with_cell_change(self):
        if isinstance(self.value, pd.DataFrame):
            column = self._cell_change["c"]  # pylint: disable=unsubscriptable-object
            index = self._cell_change["i"]  # pylint: disable=unsubscriptable-object
            new_value = self._cell_change["v"]  # pylint: disable=unsubscriptable-object
            self._pause_cds_updates = True
            self.value.at[index, column] = new_value
            self.param.trigger("value")
            self._pause_cds_updates = False

    def stream(self, stream_value: Union[pd.DataFrame, pd.Series, Dict], reset_index: bool = True):
        """Streams (appends) the `stream_value` provided to the existing value in an efficient
        manner.

        Args:
            stream_value (Union[pd.DataFrame, pd.Series, Dict]): The new value(s) to append to the
                existing value.
            reset_index (bool, optional): If the stream_value is a DataFrame and `reset_index` is
                True then the index of it is reset if True. Helps to keep the index unique and
                named `index`. Defaults to True.

        Raises:
            ValueError: Raised if the stream_value is not a supported type.

        Example: Stream a Series to a DataFrame

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> stream_value = pd.Series({"x": 4, "y": "d"})
        >>> tabulator.stream(stream_value)
        >>> tabulator.value.to_dict("list")
        {'x': [1, 2, 4], 'y': ['a', 'b', 'd']}

        Example: Stream a Dataframe to a Dataframe

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> stream_value = pd.DataFrame({"x": [3, 4], "y": ["c", "d"]})
        >>> tabulator.stream(stream_value)
        >>> tabulator.value.to_dict("list")
        {'x': [1, 2, 3, 4], 'y': ['a', 'b', 'c', 'd']}


        Example: Stream a Dictionary row to a DataFrame

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> stream_value = {"x": 4, "y": "d"}
        >>> tabulator.stream(stream_value)
        >>> tabulator.value.to_dict("list")
        {'x': [1, 2, 4], 'y': ['a', 'b', 'd']}

         Example: Stream a Dictionary of Columns to a Dataframe

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> stream_value = {"x": [3, 4], "y": ["c", "d"]}
        >>> tabulator.stream(stream_value)
        >>> tabulator.value.to_dict("list")
        {'x': [1, 2, 3, 4], 'y': ['a', 'b', 'c', 'd']}
        """

        if isinstance(self.value, pd.DataFrame):
            value_index_start = self.value.index.max() + 1
            if isinstance(stream_value, pd.DataFrame):
                if reset_index:
                    stream_value = stream_value.reset_index(drop=True)
                    stream_value.index += value_index_start
                self._pause_cds_updates = True
                self.value = pd.concat([self.value, stream_value])
                self._source.stream(stream_value)
                self._pause_cds_updates = False
            elif isinstance(stream_value, pd.Series):
                self._pause_cds_updates = True
                self.value.loc[value_index_start] = stream_value
                self._source.stream(stream_value)
                self.param.trigger("value")
                self._pause_cds_updates = False
            elif isinstance(stream_value, dict):
                if stream_value:
                    try:
                        stream_value = pd.DataFrame(stream_value)
                    except ValueError:
                        stream_value = pd.Series(stream_value)
                    self.stream(stream_value)
            else:
                raise ValueError("The patch value provided is not a DataFrame, Series or Dict!")
        else:
            self._pause_cds_updates = True
            self._source.stream(stream_value)
            self.param.trigger("value")
            self._pause_cds_updates = False

    def patch(self, patch_value: Union[pd.DataFrame, pd.Series, Dict]):
        """Patches (updates) the existing value with the `patch_value` in an efficient manner.

        Args:
            patch_value (Union[pd.DataFrame, pd.Series, Dict]): The value(s) to patch the
                existing value with.

        Raises:
            ValueError: Raised if the patch_value is not a supported type.



        Example: Patch a DataFrame with a Dictionary row.

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> patch_value = {"x": [(0, 3)]}
        >>> tabulator.patch(patch_value)
        >>> tabulator.value.to_dict("list")
        {'x': [3, 2], 'y': ['a', 'b']}

         Example: Patch a Dataframe with a Dictionary of Columns.

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> patch_value = {"x": [(slice(2), (3,4))], "y": [(1,'d')]}
        >>> tabulator.patch(patch_value)
        >>> tabulator.value.to_dict("list")
        {'x': [3, 4], 'y': ['a', 'd']}

        Example: Patch a DataFrame with a Series. Please note the index is used in the update

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> patch_value = pd.Series({"index": 1, "x": 4, "y": "d"})
        >>> tabulator.patch(patch_value)
        >>> tabulator.value.to_dict("list")
        {'x': [1, 4], 'y': ['a', 'd']}

        Example: Patch a Dataframe with a Dataframe. Please note the index is used in the update.

        >>> value = pd.DataFrame({"x": [1, 2], "y": ["a", "b"]})
        >>> tabulator = Tabulator(value=value)
        >>> patch_value = pd.DataFrame({"x": [3, 4], "y": ["c", "d"]})
        >>> tabulator.patch(patch_value)
        >>> tabulator.value.to_dict("list")
        {'x': [3, 4], 'y': ['c', 'd']}
        """
        if isinstance(self.value, pd.DataFrame):
            if isinstance(patch_value, pd.DataFrame):
                patch_value_dict = self._patch_dataframe_to_dict(patch_value)
                self.patch(patch_value_dict)
            elif isinstance(patch_value, pd.Series):
                patch_value_dict = self._patch_series_to_dict(patch_value)
                self.patch(patch_value_dict)
            elif isinstance(patch_value, dict):
                self._patch_from_dict(patch_value)
            else:
                raise ValueError(
                    f"""Patching a patch_value of type {type(patch_value)} is not supported.
                    Please provide a DataFrame, Series or Dict"""
                )
        else:
            if isinstance(patch_value, dict):
                self._pause_cds_updates = True
                self._source.patch(patch_value)
                self.param.trigger("value")
                self._pause_cds_updates = False
            else:
                raise ValueError(
                    f"""Patching a patch_value of type {type(patch_value)} is not supported.
                    Please provide a dict"""
                )

    @staticmethod
    def _patch_dataframe_to_dict(patch_value: pd.DataFrame) -> Dict:
        patch_value_dict: Dict[Any, Any] = {}
        for column in patch_value.columns:
            patch_value_dict[column] = []
            for index in patch_value.index:
                patch_value_dict[column].append((index, patch_value.loc[index, column]))

        return patch_value_dict

    @staticmethod
    def _patch_series_to_dict(patch_value: pd.Series) -> Dict:
        if "index" in patch_value:  # Series orient is row
            patch_value_dict = {k: [(patch_value["index"], v)] for k, v in patch_value.items()}
            patch_value_dict.pop("index")
        else:  # Series orient is column
            patch_value_dict = {patch_value.name: list(patch_value.items())}
        return patch_value_dict

    def _patch_from_dict(self, patch_value: Dict):
        self._pause_cds_updates = True
        for key, value in patch_value.items():
            for update in value:
                self.value.loc[update[0], key] = update[1]
        self._source.patch(patch_value)
        self.param.trigger("value")
        self._pause_cds_updates = False

    @classmethod
    def to_columns_configuration(
        cls, value: Union[pd.DataFrame, ColumnDataSource]
    ) -> List[Dict[str, str]]:
        """Returns a nice starter `columns` dictionary from the specified `value`.

        Args:
            value (Union[pd.DataFrame, ColumnDataSource]): The data source to transform.

        Returns:
            Dict: The columns configuration

        Example:

        >>> import pandas as pd
        >>> value = {"name": ["python", "panel"]}
        >>> df = pd.DataFrame(value)
        >>> Tabulator.to_columns_configuration(df)
        [{'title': 'Name', 'field': 'name', 'sorter': 'string', 'formatter': 'plaintext', \
'hozAlign': 'left'}]
        """
        col_conf = []
        for field in value.columns:
            dtype = str(value.dtypes[field])
            conf = cls._core(field=field, dtype=dtype)
            col_conf.append(conf)
        return col_conf

    @classmethod
    def _core(cls, field: str, dtype: str) -> Dict[str, str]:
        dtype_str = str(dtype)
        return {
            "title": cls._to_title(field),
            "field": field,
            "sorter": _SORTERS.get(dtype_str, "string"),
            "formatter": _FORMATTERS.get(dtype_str, "plaintext"),
            "hozAlign": _HOZ_ALIGNS.get(dtype_str, "left"),
        }

    @staticmethod
    def _to_title(field: str) -> str:
        return field.replace("_", " ").title()

    @staticmethod
    def config(css: Optional[str] = "default"):
        """Adds the specified css theme to pn.config.css_files

        Args:
            css (Optional[str], optional): [description]. Defaults to "default".
        """
        if css:
            href = CSS_HREFS[css]
            if href not in pn.config.css_files:
                pn.config.css_files.append(href)

    @property
    def selected_values(self) -> Union[pd.DataFrame, ColumnDataSource, None]:
        """Returns the selected rows of the data based

        Raises:
            ValueError: If the value is not of the supported type.

        Returns:
            Union[pd.DataFrame, ColumnDataSource, None]: The selected values of the same type as
                value. Based on the the current selection.
        """
        # Selection is a list of row indices. For example [0,2]
        if self.value is None:
            return None
        if isinstance(self.value, pd.DataFrame):
            return self.value.iloc[
                self.selection,
            ]
        if isinstance(self.value, ColumnDataSource):
            # I could not find a direct way to get a selected ColumnDataSource
            selected_data = self.value.to_df().iloc[
                self.selection,
            ]
            return ColumnDataSource(selected_data)
        raise ValueError("The value is not of a supported type!")

    @param.depends("selection", watch=True)
    def _update_source_selected_indices(self, *_):
        self._source.selected.indices = self.selection

    def _get_model(self, doc, root=None, parent=None, comm=None):
        model = super()._get_model(doc=doc, root=root, parent=parent, comm=comm)
        if root is None:
            root = model
        self._link_props(model.source.selected, ["indices"], doc, root, comm)
        return model

    def _process_events(self, events):
        if "indices" in events:
            self.selection = events.pop("indices")
        super()._process_events(events)


class TabulatorStylesheet(pn.pane.HTML):
    """The TabulatorStyleSheet provides methods to dynamically change the (css) style of the
    Tabulator widget"""

    theme = param.ObjectSelector(default="site", objects=sorted(list(CSS_HREFS.keys())))

    # In order to not be selected by the `pn.panel` selection process
    # Cf. https://github.com/holoviz/panel/issues/1494#issuecomment-663219654
    priority = 0
    # The _rename dict is used to keep track of Panel parameters to sync to Bokeh properties.
    # As value is not a property on the Bokeh model we should set it to None
    _rename = {
        **pn.pane.HTML._rename,
        "theme": None,
    }

    def __init__(self, **params):
        params["height"] = 0
        params["width"] = 0
        params["sizing_mode"] = "fixed"
        params["margin"] = 0
        super().__init__(**params)

        self._update_object_from_parameters()

    # Don't name the function
    # `_update`, `_update_object`, `_update_model` or `_update_pane`
    # as this will override a function in the parent class.
    @param.depends("theme", watch=True)
    def _update_object_from_parameters(self, *_):
        href = CSS_HREFS[self.theme]
        self.object = f'<link rel="stylesheet" href="{href}">'

    def __repr__(self, depth=0):  # pylint: disable=unused-argument
        return f"Tabulator({self.name})"

    def __str__(self):
        return f"Tabulator({self.name})"
